"""
Base class for trainer initializers
"""
from abc import ABC, abstractmethod
import ray

from data.dataset_util import split_dataset_for_classification


class TrainerInitializerBase(ABC):

    def __init__(self, storage_manager, model, config):
        super().__init__()
        self.storage_manager = storage_manager
        self.model = model
        self.model_name = model.model_name
        self.config = config

    @abstractmethod
    def model_init(self):
        raise NotImplementedError('need to implement model_init()')

    def get_train_and_eval_datasets(self, model):
        dataset = self.dataset_init(model, is_eval=False)
        train_dataset, eval_dataset = split_dataset_for_classification(dataset, 0.9) # not using a validation dataset right now because our dataset is too small for classes
        return train_dataset, eval_dataset

    def dataset_init(self, model, is_eval=False):
        return self.do_dataset_init(model, is_eval=is_eval)

    @abstractmethod
    def do_dataset_init(self, model, is_eval=False):
        raise NotImplementedError('need to implement do_dataset_init()')

    def convert_to_ray_dataset(self, dataset):
        return ray.data.from_items(dataset)

    @abstractmethod
    def trainer_args_init(self, model):
        raise NotImplementedError('need to implement trainer_args_init()')

    def preprocessor_init(self, model):
        return None

    @abstractmethod
    def data_collator_init(self, model):
        raise NotImplementedError('need to implement data_collator_init()')

    def compute_metrics_init(self):
        return None

    def compute_objective_init(self):
        return None

    @abstractmethod
    def trainer_init(self, model, args, train_dataset, eval_dataset, scaling_config):
        raise NotImplementedError('need to implement trainer_init()')

    @abstractmethod
    def trainer_init_per_worker(self, train_dataset, eval_dataset, **trainer_init_config):
        raise NotImplementedError('need to implement trainer_init_per_worker()')

    @abstractmethod
    def update_model_with_best_checkpoint(self, model, config, checkpoints, default_eval_metric):
        raise NotImplementedError('need to implement trainer_init()')



"""
Convenience function for getting info about the current OS platform
"""
import os
import sys

import torch


def get_device_type():
    """returns the device type as a string"""
    # noinspection PyUnresolvedReferences
    if torch.cuda.is_available():
        return 'cuda'
    elif torch.backends.mps.is_available():
        return 'mps'
    else:
        return 'cpu'


def get_cuda_device_count():
    return torch.cuda.device_count()


def get_cpu_device_count():
    return os.cpu_count()


def is_linux():
    return sys.platform == 'linux' or sys.platform == 'linux2'


def is_mac():
    return sys.platform == 'darwin'


def is_windows():
    return sys.platform == 'win32'


def expand_user_home_path(path, user, platform):
    if path.startswith('~'):
        if platform == 'darwin':
            path = path.replace('~/', f'/Users/{user}/')
        elif platform == 'windows':
            path = path.replace('~/', f'C:/Users/{user}/')
        else:
            path = path.replace('~/', f'/home/{user}/')
    else:
        if platform == 'darwin':
            if path.startswith('/home/'):
                path = path.replace('/home/', '/Users/')
            elif path.startswith('C:/Users/'):
                path = path.replace('C:/Users/', '/Users/')
        elif platform == 'windows':
            if path.startswith('/Users/'):
                path = path.replace('/Users/', 'C:/Users/')
            elif path.startswith('/home/'):
                path = path.replace('/home/', 'C:/Users/')
        else:
            if path.startswith('/Users/'):
                path = path.replace('/Users/', '/home/')
            elif path.startswith('C:/Users/'):
                path = path.replace('C:/Users/', '/home/')
    return path



import os
import sys

import numpy as np
from ray import logger


def monkey_patch_cloudpickle_fast():
    """This can be useful if you need to debug cloud-pickling issues"""
    from ray.cloudpickle.cloudpickle_fast import CloudPickler
    from pickle import Pickler
    import pickle

    def dump(self, obj):
        try:
            logger.info(f"DUMPING {obj}")
            return Pickler.dump(self, obj)
        except RuntimeError as e:
            if "recursion" in e.args[0]:
                msg = (
                    "Could not pickle object as excessively deep recursion "
                    "required."
                )
                raise pickle.PicklingError(msg) from e
            else:
                raise

    CloudPickler.dump = dump


def monkey_patch_base_trainer_to_enable_syncing_after_training():
    """Monkey-patch ray.train.base_trainer.BaseTrainer so we can't sync the checkpoints before retrieving the grid results so that the checkpoints are available locally."""

    from ray.train.base_trainer import BaseTrainer
    from ray.air import Result
    from ray.train.base_trainer import TrainingFailedError
    from ray.util import PublicAPI

    @PublicAPI(stability="beta")
    def fit(self, syncer=None) -> Result:
        """Runs training.

        Returns:
            A Result object containing the training result.

        Raises:
            TrainingFailedError: If any failures during the execution of
            ``self.as_trainable()``.
        """
        from ray.tune.tuner import Tuner
        from ray.tune.error import TuneError

        trainable = self.as_trainable()

        tuner = Tuner(trainable=trainable, run_config=self.run_config)
        result_grid = tuner.fit()
        assert len(result_grid) == 1
        if syncer is not None:
            syncer.sync_from_ray_worker_to_driver()
        try:
            result = result_grid[0]
            if result.error:
                raise result.error
        except TuneError as e:
            raise TrainingFailedError from e
        return result

    BaseTrainer.fit = fit


def monkey_patch_huggingface_utils_to_process_datasets_for_gpt2():
    """Monkey-patch huggingface.utils to convert the NumPy arrays into labelled examples for GPT2 training"""
    from ray.train.huggingface._huggingface_utils import RayDatasetHFIterable

    def __iter__(self):
        for row in self.generate_examples_fn(**self.kwargs):
            example = (0, {'input_ids': row[0:len(row)-1].astype(np.int64),
                           'labels': row[1:len(row)].astype(np.int64)}
                       )
            yield example

    RayDatasetHFIterable.__iter__ = __iter__


def monkey_patch_trainable_util_to_fix_checkpoint_paths():
    """Monkey-patch ray.tune.trainable.util.TrainableUtil to fix the checkpoint path in the home directory so the path is correct when transferred from the Ray worker to the driver"""
    from ray.tune.trainable.util import TrainableUtil

    def find_checkpoint_dir(checkpoint_path):
        """Returns the directory containing the checkpoint path.

        Raises:
            FileNotFoundError if the directory is not found.
        """
        if sys.platform == 'darwin':
            checkpoint_path = checkpoint_path.replace('/home', '/Users')
        elif sys.platform == 'win32':
            checkpoint_path = checkpoint_path.replace('/home', 'C:/Users')
        if not os.path.exists(checkpoint_path):
            raise FileNotFoundError("Path does not exist", checkpoint_path)
        if os.path.isdir(checkpoint_path):
            checkpoint_dir = checkpoint_path
        else:
            checkpoint_dir = os.path.dirname(checkpoint_path)
        while checkpoint_dir != os.path.dirname(checkpoint_dir):
            if os.path.exists(os.path.join(checkpoint_dir, ".is_checkpoint")):
                break
            checkpoint_dir = os.path.dirname(checkpoint_dir)
        else:
            raise FileNotFoundError(
                "Checkpoint directory not found for {}".format(checkpoint_path)
            )
        return os.path.normpath(checkpoint_dir)

    TrainableUtil.find_checkpoint_dir = find_checkpoint_dir



import torch
from transformers import AutoModelForCausalLM, GPT2Config, GPT2Model, GPT2Tokenizer, pipeline

from config import config
from models.model_base import ModelBase

PRE_TRAINED_MODEL = 'gpt2'


class GPT2(ModelBase):

    def __init__(self, storage_manager, model_name, pipeline_name, use_trained_model=False, is_inference_mode=False):
        self.pipeline_name = pipeline_name
        self.pipeline = None
        self.tokenizer = None
        super(GPT2, self).__init__(model_name, storage_manager, use_trained_model, is_inference_mode)

    def _do_create_model_config(self):
        return GPT2Config.from_pretrained(PRE_TRAINED_MODEL)

    def _do_create_model(self, model_config):
        # for text generation, we need a GPT2 model with a language model head: the easiest way to instantiate the
        # correct model is to grab it from the pipeline
        if self.pipeline_name is not None:
            model = AutoModelForCausalLM.from_pretrained(PRE_TRAINED_MODEL,
                                                         config=model_config,
                                                         _from_pipeline=self.pipeline_name,
                                                         ignore_mismatched_sizes=True)
            self.tokenizer = GPT2Tokenizer.from_pretrained(PRE_TRAINED_MODEL, config=model_config)
            self.pipeline = pipeline(self.pipeline_name, model=model, tokenizer=self.tokenizer)
            return model
        else:
            self.tokenizer = GPT2Tokenizer.from_pretrained(PRE_TRAINED_MODEL, config=model_config)
            return GPT2Model.from_pretrained(PRE_TRAINED_MODEL,
                                             config=model_config,
                                             ignore_mismatched_sizes=True)

    def _do_load_model(self, model_config):
        self.tokenizer = GPT2Tokenizer.from_pretrained(PRE_TRAINED_MODEL, config=model_config)
        return GPT2Model.from_pretrained(self.get_model_path())

    def get_save_model_in_folder(self):
        return True

    def save_model(self):
        if self.model is None:
            return
        self.model.save_pretrained(self.get_model_path())

    def tokenize(self, prompt_text, return_tensors=False):
        encoded_input = self.tokenizer(prompt_text,
                                       padding=False,
                                       add_special_tokens=False,
                                       return_tensors=return_tensors).data
        if return_tensors == 'pt':
            for key in encoded_input:
                encoded_input[key] = encoded_input[key].to(self.get_device_type())
        return encoded_input

    def forward(
            self,
            input_ids=None,
            past_key_values=None,
            attention_mask=None,
            token_type_ids=None,
            position_ids=None,
            head_mask=None,
            inputs_embeds=None,
            labels=None,
            use_cache=None,
            output_attentions=None,
            output_hidden_states=None,
            return_dict=None):
        return self.get_model()(input_ids=input_ids,
                                past_key_values=past_key_values,
                                attention_mask=attention_mask,
                                token_type_ids=token_type_ids,
                                position_ids=position_ids,
                                head_mask=head_mask,
                                inputs_embeds=inputs_embeds,
                                labels=labels,
                                use_cache=use_cache,
                                output_attentions=output_attentions,
                                output_hidden_states=output_hidden_states,
                                return_dict=return_dict)

    def _do_train_model(self):
        raise Exception("Use the Trainer for this model instead of calling train_model()")

    def _do_update_model(self, message_ids):
        pass

    def set_eval_mode(self):
        super().set_eval_mode()

    def validate_model(self):
        pass

    @torch.no_grad()
    def predict(self, input):
        self.set_eval_mode()
        # extracted from TextGenerationPipeline in transformers/pipelines.py so I can play with the inner workings
        if self.pipeline_name == 'text-generation':
            encoded_input = self.tokenize(input, return_tensors=config.framework)
            input_ids = encoded_input["input_ids"]
            attention_mask = encoded_input.get("attention_mask", None)
            # Allow empty prompts
            if input_ids.shape[1] == 0:
                input_ids = None
                attention_mask = None
                in_b = 1
            else:
                in_b = input_ids.shape[0]
            model_outputs = self.model.generate(input_ids=input_ids,
                                                attention_mask=attention_mask,
                                                max_new_tokens=config.response_max_tokens)
            out_b = model_outputs.shape[0]
            generated_sequence = model_outputs.reshape(in_b, out_b // in_b, *model_outputs.shape[1:])
            model_outputs = {'generated_sequence': generated_sequence, 'input_ids': input_ids, 'prompt_text': input}
            response = self.postprocess(model_outputs)
            return response
        elif self.pipeline is not None:
            return self.pipeline(input)
        else:
            raise Exception("No pipeline defined")

    @torch.no_grad()
    def predict_batch(self, prompt_texts):
        """returns a list of predictions (predicted_class_name, probability, probabilities) for the input messages"""
        self.set_eval_mode()
        encoded_input = self.tokenize(prompt_texts, return_tensors=config.framework)
        input_ids = encoded_input["input_ids"]
        attention_mask = encoded_input.get("attention_mask", None)
        # Allow empty prompts
        if input_ids.shape[1] == 0:
            input_ids = None
            attention_mask = None
            in_b = 1
        else:
            in_b = input_ids.shape[0]
        model_outputs = self.generate(input_ids=input_ids, attention_mask=attention_mask)
        out_b = model_outputs.shape[0]
        generated_sequence = model_outputs.reshape(in_b, out_b // in_b, *model_outputs.shape[1:])
        response = self.postprocess(generated_sequence)
        return response

    def postprocess(self, model_outputs):
        generated_sequence = model_outputs['generated_sequence'][0].numpy().tolist()
        input_ids = model_outputs["input_ids"]
        prompt_text = model_outputs["prompt_text"]
        responses = []
        for sequence in generated_sequence:
            # decode text
            text = self.tokenizer.decode(
                sequence,
                skip_special_tokens=True,
                clean_up_tokenization_spaces=True,
            )
            if input_ids is None:
                prompt_length = 0
            else:
                prompt_length = len(
                    self.tokenizer.decode(
                        input_ids[0],
                        skip_special_tokens=True,
                        clean_up_tokenization_spaces=True,
                    )
                )
            text = text[prompt_length:]
            responses.append(text)
        return len(responses) == 1 and responses[0] or responses



import torch
from torch.utils.data import random_split

from config import config


def split_dataset_random(dataset, split_ratio):
    """Split dataset into two datasets according to split ratio"""
    split_index = int(len(dataset) * split_ratio)
    return random_split(dataset,
                        [split_index, len(dataset) - split_index],
                        generator=torch.Generator().manual_seed(config.seed))


def split_dataset_for_classification(dataset, split_ratio):
    """Split dataset into two datasets according to split ratio and stratify by class"""
    examples_for_classes_in_dataset = {}
    for example in dataset:
        label = example['label']
        if label not in examples_for_classes_in_dataset:
            examples_for_classes_in_dataset[label] = []
        examples_for_classes_in_dataset[label].append(example)
    train_dataset = []
    eval_dataset = []
    for example_class in examples_for_classes_in_dataset.keys():
        examples = examples_for_classes_in_dataset[example_class]
        num_examples = len(examples)
        if num_examples == 0:
            continue
        elif num_examples == 1:
            train_dataset.append(examples[0])
        elif num_examples == 2:
            train_dataset.append(examples[0])
            eval_dataset.append(examples[1])
        else:
            train_examples, test_examples = split_dataset_random(examples, split_ratio)
            train_dataset.extend(train_examples)
            eval_dataset.extend(test_examples)
    return train_dataset, eval_dataset



"""
Dataset for use with GPT2 model.
"""
import numpy as np
from torch.utils.data import Dataset

from log import log


class GPT2Dataset(Dataset):
    """
    This dataset is unique in that the examples are permutations of the same underlying text. It would be very inefficient
    to generate all the examples locally and then copy them over the network. Instead, we just send the underlying text and
    generate the examples on the fly from the monkey-patched RayDatasetHFIterable.
    """

    def __init__(self, model, use_memmap=False, is_eval=False):
        super().__init__()
        self.model = model
        self.dataset_name = is_eval and 'val' or 'train'
        self.block_size = model.model.config.block_size
        if use_memmap:
            self.data = np.memmap(self.get_file_path(),
                                  dtype=np.uint16,
                                  mode='r')
        else:
            self.data = np.fromfile(self.get_file_path(),
                                    dtype=np.uint16)
        log.info(f'loaded {len(self.data)} examples')

    def get_file_path(self):
        return f'{self.model.storage_manager.get_data_dir()}/{self.model.model_name}/{self.dataset_name}.bin'

    def __len__(self):
        return self.get_num_examples()

    def __getitem__(self, index):
        if index >= len(self):
            raise IndexError('index out of range')
        x = self.data[index:index+self.block_size].astype(np.int64)
        y = self.data[index+1:index+1+self.block_size].astype(np.int64)
        return {'input_ids': x, 'labels': y}

    def get_data_for_ray_dataset(self):
        # get the data as a numpy array to allow for efficient transfer to the remote workers
        return np.array([self.data[i:i+self.block_size+1] for i in range(self.get_num_examples())]) # add 1 to data length so we can create the labels later (data[x + 1:] is the label for data[x:])

    def get_num_examples(self):
        """You can edit to return a smaller number if you are running out of memory when trying to train on this dataset"""
        return len(self.data) - self.block_size



"""
Manager for managing directories and files for the project.
"""
import os
import shutil

from config import BASE_DIR
from log import log


class StorageManager:

    def get_config_dir(self):
        return BASE_DIR + '/config'

    def get_data_base_dir(self):
        return BASE_DIR + '/data'

    def get_data_dir(self):
        return f'{self.get_data_base_dir()}'

    def get_logs_dir(self):
        return BASE_DIR + '/logs'

    def get_models_dir(self):
        return BASE_DIR + '/models'

    def get_runs_dir(self):
        return BASE_DIR + '/runs'

    def get_src_dir(self):
        return BASE_DIR + '/src'

    def clean_for_training(self):
        self._delete_training_checkpoint_dirs()
        self._delete_training_logs_dirs()

    def _delete_training_checkpoint_dirs(self):
        if not os.path.exists(self.get_runs_dir()):
            return
        for dir_name in os.listdir(self.get_runs_dir()):
            if dir_name.startswith('checkpoint-'):
                self._delete_dir(f'{self.get_runs_dir()}/{dir_name}')

    def _delete_training_logs_dirs(self):
        if not os.path.exists(self.get_logs_dir()):
            return
        for filename in os.listdir(self.get_logs_dir()):
            if os.path.isdir(f'{self.get_logs_dir()}/{filename}'):
                self._delete_dir(f'{self.get_logs_dir()}/{filename}')
            if filename.startswith('events.out'):
                os.remove(f'{self.get_logs_dir()}/{filename}')

    def _delete_dir(self, dir_path):
        if not os.path.exists(dir_path):
            return
        log.info(f'deleting {dir_path}...')
        try:
            shutil.rmtree(dir_path)
        except OSError:
            os.remove(dir_path)

    def copy_file(self, filename, src_dir, dst_dir):
        log.info(f'copying {src_dir} to {dst_dir}')
        if os.path.exists(f'{dst_dir}/{filename}'):
            os.remove(f'{dst_dir}/{filename}')
        shutil.copy(f'{src_dir}/{filename}', dst_dir)



"""
Trainer initializer for the GPT model.
"""
import ray
from transformers import DefaultDataCollator

from config import config
from data.gpt2_dataset import GPT2Dataset
from models.gpt2 import GPT2
from ray_quickstart.monkey_patch import monkey_patch_huggingface_utils_to_process_datasets_for_gpt2
from training.huggingface_trainer_initializer_base import HuggingFaceTrainerInitializerBase


class GPT2TrainerInitializer(HuggingFaceTrainerInitializerBase):

    def __init__(self, storage_manager, model):
        super().__init__(storage_manager, model, config)

    def get_pipeline_name(self):
        if self.model is not None:
            return self.model.pipeline_name
        else:
            return 'text-generation'

    def model_init(self):
        model = GPT2(self.storage_manager, self.model_name, self.get_pipeline_name(), False, False)
        model.load_or_create_model()
        model.set_train_mode()
        return model

    def get_train_and_eval_datasets(self, model):
        train_dataset = self.dataset_init(model, is_eval=False)
        eval_dataset = self.dataset_init(model, is_eval=True)
        return train_dataset, eval_dataset

    def do_dataset_init(self, model, is_eval=False):
        return GPT2Dataset(model, is_eval=is_eval)

    def convert_to_ray_dataset(self, dataset):
        return ray.data.from_numpy(dataset.get_data_for_ray_dataset())

    def data_collator_init(self, model):
        return DefaultDataCollator()

    def trainer_init_per_worker(self, train_dataset, eval_dataset, **trainer_init_config):
        monkey_patch_huggingface_utils_to_process_datasets_for_gpt2()
        return super().trainer_init_per_worker(train_dataset, eval_dataset, **trainer_init_config)



import logging
import logging.config
import os
import pathlib


LOGS_DIR = str(pathlib.Path(__file__).parents[2]).replace('\\' , '/') + '/logs'


def setup_logging():
    if not os.path.exists(LOGS_DIR):
        os.mkdir(LOGS_DIR)
    file_filename = f'{LOGS_DIR}/ray_quickstart.log'
    file_write_mode = 'w'
    file_log_level = 'DEBUG'
    file_log_formatter = 'file'
    config_dir = str(pathlib.Path(__file__).parents[2]).replace('\\' , '/') + '/config'
    if os.path.exists(f'{config_dir}/logging.ini'):
        logging.config.fileConfig(f'{config_dir}/logging.ini',
                                  defaults={'file_filename': file_filename,
                                            'file_write_mode': file_write_mode,
                                            'file_log_level': file_log_level,
                                            'file_log_formatter': file_log_formatter},
                                  disable_existing_loggers=True)

    # Tensorflow logging
    logging.getLogger('h5py').setLevel(logging.INFO)
    logging.getLogger('tensorflow').setLevel(logging.INFO)

    # transformers logging
    logging.getLogger('transformers.configuration_utils').setLevel(logging.WARN)

    logger = logging.getLogger('app')
    logger.setLevel(logging.INFO)

    return logger


def truncate_logs():
    if os.path.exists(LOGS_DIR + '/ray_quickstart.log'):
        print('truncating log files')
        with open(f'{LOGS_DIR}/ray_quickstart.log', 'w') as f:
            f.truncate(0)


log = setup_logging()



"""
Base class for trainer implementations.
"""
import ray
from ray.air import ScalingConfig

from config import BASE_DIR, CONFIG_DIR, SRC_DIR
from data.dataset_util import split_dataset_random
from log import log
from ray_quickstart import initialize_ray_with_syncer
from ray_quickstart import initialize_ray
from ray_quickstart.util.platform import get_cpu_device_count


def train(trainer_initializer, config):
    model = trainer_initializer.model
    model.load_or_create_model()
    model.is_training = True
    log.info(f'training {model.model_name} model...')
    train_dataset, eval_dataset = trainer_initializer.get_train_and_eval_datasets(model)
    if config.get_run_on_ray_cluster():
        syncer = initialize_ray_with_syncer(BASE_DIR,
                                            SRC_DIR,
                                            f'{CONFIG_DIR}/ray_config.yaml',
                                            config.trial_results_dir)
        log.info('training using ray cluster...')
        ray_train_dataset = trainer_initializer.convert_to_ray_dataset(train_dataset)
        ray_eval_dataset = trainer_initializer.convert_to_ray_dataset(eval_dataset)
        scaling_config = create_scaling_config()
        args = trainer_initializer.trainer_args_init(model)
        trainer = trainer_initializer.trainer_init(model,
                                                   args,
                                                   ray_train_dataset,
                                                   ray_eval_dataset,
                                                   scaling_config)
        result = trainer.fit(syncer)
        checkpoint = len(result.best_checkpoints) > 0
        if checkpoint is None:
            log.info('no best checkpoint found after training using ray cluster')
        else:
            trainer_initializer.update_model_with_best_checkpoint(model, config, result.best_checkpoints, args.metric_for_best_model)
    else:
        log.info('training using local computer...')
        model.set_train_mode()
        trainer_init_config = {
            'model': model,
            'args': trainer_initializer.trainer_args_init(model),
            'data_collator': trainer_initializer.data_collator_init(model),
            'compute_metrics': trainer_initializer.compute_metrics_init()
        }
        trainer = trainer_initializer.trainer_init_per_worker(train_dataset, eval_dataset, **trainer_init_config)
        result = trainer.train()
        log.info(f'Best checkpoint metrics: {result.metrics}')
        model = trainer.model
        model.save_model()
    log.info(f'finished training {model.model_name} model')
    return model


def tune_hyperparameters(trainer_initializer, config):
    initialize_ray(SRC_DIR, f'{CONFIG_DIR}/ray_config.yaml')
    model = trainer_initializer.model_init()
    evaluation_strategy = 'epoch' # 'no', 'steps', or 'epoch'
    args = trainer_initializer.trainer_args_init(model,
                                                 tensorboard_logging_strategy='no',
                                                 save_strategy='no',
                                                 evaluation_strategy=evaluation_strategy,
                                                 disable_tqdm=True)
    scaling_config = create_scaling_config()
    dataset = trainer_initializer.dataset_init(model, is_eval=False)
    train_dataset, eval_dataset = split_dataset_random(dataset, 0.9) # not using a validation dataset right now because our dataset is too small for classes
    trainer_init_config = {
        'model_init': trainer_initializer.model_init,
        'args': args,
        'data_collator': trainer_initializer.data_collator_init(model),
        'compute_metrics': trainer_initializer.compute_metrics_init()
    }
    trainer = trainer_initializer.trainer_init_per_worker(train_dataset, eval_dataset, **trainer_init_config)
    search_space = {
        'learning_rate': ray.tune.loguniform(1e-5, 1e-3),
        'per_device_train_batch_size': ray.tune.choice([4, 8, 16, 32]),
        'per_device_eval_batch_size': 32,
        'num_train_epochs': ray.tune.choice([5, 6, 7, 8, 9]),
        'weight_decay': ray.tune.loguniform(1e-3, 1e-1)
    }
    num_train_epochs_search_space = {
        'learning_rate': 6e-5,
        'per_device_train_batch_size': 32,
        'per_device_eval_batch_size': 32,
        'num_train_epochs': ray.tune.choice([20, 25, 30, 35, 40, 45, 50]),
        'weight_decay': 1e-2
    }
    from ray.tune.schedulers import ASHAScheduler
    from ray.tune.search.hyperopt import HyperOptSearch
    best_trial = trainer.hyperparameter_search(
        hp_space=lambda _:search_space,
        compute_objective=trainer_initializer.compute_objective_init(),
        n_trials=10,
        direction='maximize',
        backend='ray',
        resources_per_trial={'cpu': scaling_config.resources_per_worker['CPU'],
                             'gpu': scaling_config.resources_per_worker['GPU']},
        search_alg=HyperOptSearch(metric='objective', mode='max'),
        scheduler=ASHAScheduler(metric='objective', mode='max')
    )
    log.info(f'best trial: {best_trial}')


def create_scaling_config():
    use_gpu = True
    num_trainer_cpus = 4
    if use_gpu:
        num_cpus = 0
        num_gpus = 1
    else:
        num_cpus = get_cpu_device_count() - num_trainer_cpus - 1
        num_gpus = 0
    scaling_config = ScalingConfig(num_workers=1,
                                   use_gpu=use_gpu,
                                   trainer_resources={'CPU': num_trainer_cpus, 'GPU': 0},
                                   resources_per_worker={'CPU': num_cpus, 'GPU': num_gpus})
    return scaling_config



"""
Base for ML models
"""
from abc import abstractmethod
import numpy as np
import os
import torch
from torch.nn import Module
from typing import final
import yaml

from config import config
from log import log


class ModelBase(Module):

    def __init__(self, model_name, storage_manager, use_trained_model, is_inference_mode=False):
        super().__init__()
        self.model_name = model_name
        self.storage_manager = storage_manager
        self.use_trained_model = use_trained_model
        self.is_inference_mode = is_inference_mode
        self.model = None
        self.is_training = False
        self.training_validation_accuracy = None
        self.models_dir_override = None # set to load model from checkpoint

    def get_device_type(self):
        return self.get_model().device

    def is_persistent_model(self):
        return True

    def set_models_dir(self, models_dir):
        self.models_dir_override = models_dir

    def get_model(self):
        if self.model is None:
            self.load_or_create_model()
        return self.model

    def get_model_path(self):
        return self._get_model_path(self.model_name)

    def _get_model_path(self, model_name):
        if self.models_dir_override is not None:
            models_dir = self.models_dir_override
        else:
            models_dir = self.storage_manager.get_models_dir()
        if self.get_save_model_in_folder():
            return f'{models_dir}/{model_name}_model'
        else:
            return f'{models_dir}/{self.get_model_framework()}_{model_name}_model'

    def get_model_framework(self):
        return config.framework == 'pt' and 'pytorch' or 'tensorflow'

    @final
    def load_or_create_model(self):
        if self.model is not None:
            return
        model_config = self.get_model_config()
        if self.is_persistent_model and self.use_trained_model:
            if os.path.exists(self.get_model_path() + (not self.get_save_model_in_folder() and '.p' or '')):
                log.info(f'loading model from {self.get_model_path()}')
                self.model = self._do_load_model(model_config)
            else:
                self.use_trained_model = False
        if self.model is None:
            log.info(f'creating {self.model_name} model')
            self.model = self._do_create_model(model_config)
        if self.is_inference_mode:
            self.to(torch.device(config.device_type))
        return self.model

    def load_from_checkpoint(self, checkpoint_dir):
        self.models_dir_override = checkpoint_dir
        model_config = self.get_model_config()
        self.model = self._do_load_model(model_config)
        self.models_dir_override = None
        self.save_model()

    def get_model_config(self):
        model_config = self._do_create_model_config()
        model_config_overrides = self.load_model_config()
        if model_config_overrides is not None:
            model_config.update(model_config_overrides)
        return model_config

    def load_model_config(self, model_config_name=None):
        model_config_file_path = self._get_model_config_file_path(model_config_name)
        if os.path.exists(model_config_file_path):
            with open(model_config_file_path, 'r') as f:
                model_config = yaml.load(f, Loader=yaml.FullLoader)
                return model_config
        return None

    def save_model_config(self, model_config, model_config_name=None):
        model_config_file_path = self._get_model_config_file_path(model_config_name)
        with open(model_config_file_path, 'w') as f:
            yaml.dump(model_config, f, default_flow_style=False)

    def _get_model_config_file_path(self, model_config_name=None):
        if model_config_name is not None:
            return f'{self.storage_manager.get_config_dir()}/{self.model_name}/model_{model_config_name}.yaml'
        else:
            return f'{self.storage_manager.get_config_dir()}/{self.model_name}/model.yaml'

    @abstractmethod
    def _do_create_model_config(self):
        raise NotImplementedError('need to implement do_create_model_config()')

    @abstractmethod
    def _do_create_model(self, model_config):
        raise NotImplementedError('need to implement do_load_model()')

    def _do_load_model(self, model_config):
        return torch.load(self.get_model_path(), map_location=torch.device(config.device_type))

    def get_save_model_in_folder(self):
        """whether the model is saved in a folder or in a single file"""
        return False

    def save_model(self):
        if self.is_persistent_model():
            self.model.save(self.get_model_path())

    def to(self, device_type):
        if self.get_device_type() != device_type:
            log.info(f'converting {self.model_name} model to device {device_type}')
            self.model.to(device_type)
        return self

    def load_training_params(self, train_params_name=None):
        training_params_file_path = self._get_train_params_file_path(train_params_name)
        if os.path.exists(training_params_file_path):
            with open(training_params_file_path, 'r') as f:
                training_params = yaml.load(f, Loader=yaml.FullLoader)
                return training_params
        return None

    def save_training_params(self, training_params, train_params_name=None):
        training_params_file_path = self._get_train_params_file_path(train_params_name)
        with open(training_params_file_path, 'w') as f:
            yaml.dump(training_params, f, default_flow_style=False)

    def _get_train_params_file_path(self, training_params_name=None):
        if training_params_name is not None:
            return f'{self.storage_manager.get_config_dir()}/{self.model_name}/train_{training_params_name}.yaml'
        else:
            return f'{self.storage_manager.get_config_dir()}/{self.model_name}/train.yaml'

    def set_train_mode(self):
        self.get_model().train()

    @abstractmethod
    def forward(self, *args, **kwargs):
        raise NotImplementedError('need to implement forward()')

    @final
    def train_model(self):
        if self.use_trained_model:
            log.info('not training because use_trained_model is set to True')
            return
        log.info(f'training {self.model_name} model...')
        self.load_or_create_model()
        self.set_train_mode()
        self.is_training = True
        self._do_train_model()
        self.save_model()
        log.info(f'finished training {self.model_name} model')

    @abstractmethod
    def _do_train_model(self):
        raise NotImplementedError('need to implement do_train_model()')

    def get_training_validation_accuracy(self):
        return self.training_validation_accuracy

    @final
    def update_model(self, message_ids):
        if self.is_persistent_model() and not self.use_trained_model:
            log.info('not updating because use_trained_model is set to False: you might want to call train() instead')
            return
        if self.is_persistent_model():
            log.info(f'updating {self.model_name} model...')
        self.load_or_create_model()
        self.is_training = True
        self._do_update_model(message_ids)
        self.save_model()
        if self.is_persistent_model():
            log.info(f'finished updating {self.model_name} model')

    @abstractmethod
    def _do_update_model(self, message_ids):
        raise NotImplementedError('need to implement do_update_model()')

    def set_eval_mode(self):
        self.get_model().eval()

    @abstractmethod
    def predict(self, *args, **kwargs):
        raise NotImplementedError('need to implement predict()')

    @abstractmethod
    def predict_batch(self, *args, **kwargs):
        raise NotImplementedError('need to implement predict_batch()')



import pathlib

import numpy as np
import torch

from ray_quickstart.init import platform

BASE_DIR = str(pathlib.Path(__file__).parents[2]).replace('\\' , '/')
CONFIG_DIR = BASE_DIR + '/config'
RUNS_DIR = BASE_DIR + '/runs'
SRC_DIR = BASE_DIR + '/src'


class Config:

    def __init__(self):
        super().__init__()

        self.seed = 1234
        np.random.seed(self.seed)
        torch.manual_seed(self.seed)

        self.framework = 'pt' # 'pt' or 'tf'
        self.device_type = platform.get_device_type()

        self.trial_results_dir = '~/ray_results'

        self.run_on_ray_cluster = platform.is_mac()

    def get_run_on_ray_cluster(self):
        return self.run_on_ray_cluster


config = Config()



"""
Utilities for working with Ray.
"""
import os
import shutil
import subprocess

import ray
from ray import logger
import yaml

from ray_quickstart.monkey_patch import monkey_patch_base_trainer_to_enable_syncing_after_training, \
    monkey_patch_trainable_util_to_fix_checkpoint_paths
from ray_quickstart.rsync_syncer import RsyncSyncer
from ray_quickstart.util import platform
from ray_quickstart.util.platform import expand_user_home_path


def initialize_ray_with_syncer(base_dir,
                               src_dir,
                               ray_config_file_path,
                               trial_results_dir,
                               success_callback=None,
                               clean_trial_results_dir_at_start=True):
    """
    :param base_dir: The base directory of your project.
    :param src_dir: The base directory for your Python source code
    :param ray_config_file_path: The path to the Ray config file.
    :param trial_results_dir The directory where the Ray trial results are stored. This directory needs to have the same
           path on your local computer and at the remote computer. If the directory is in your user home directory, we
           will do the platform-dependent path expansion for you.
    :param success_callback: Callback to call when Ray is successfully initialized.
    :param clean_trial_results_dir_at_start: Whether to clean the trial results directory on your local computer and the remote computer at the start of the experiment.
    :return: an RsyncSyncer object that needs to be called after training to sync the checkpoints from the remote computer to the local computer.
    """
    ray_config = load_ray_config(ray_config_file_path)
    driver_user = ray_config['driver']['user']
    driver_private_key_file = ray_config['driver']['private_key_file']
    worker_user = ray_config['worker']['user']
    worker_hostname_or_ip_address = ray_config['worker']['hostname_or_ip_address']
    worker_ssh_port = ray_config['worker']['ssh_port']
    worker_platform = ray_config['worker']['platform']
    worker_setup_commands = ray_config['worker']['setup_commands']
    syncer = RsyncSyncer(driver_user,
                         driver_private_key_file,
                         worker_user,
                         worker_hostname_or_ip_address,
                         worker_ssh_port,
                         worker_platform,
                         trial_results_dir)
    if ray.is_initialized():
        return syncer
    monkey_patch_base_trainer_to_enable_syncing_after_training()
    try:
        if clean_trial_results_dir_at_start:
            clean_trial_results_dir(syncer, trial_results_dir)
        configure_remote_ray_runtime_environment(base_dir,
                                                 driver_private_key_file,
                                                 worker_user,
                                                 worker_hostname_or_ip_address,
                                                 worker_ssh_port,
                                                 worker_platform,
                                                 worker_setup_commands)
        initialize_ray(src_dir, ray_config_file_path, ray_config, success_callback)
        return syncer
    except ConnectionError:
        if platform.is_windows() and os.path.exists(f'{base_dir}/scripts/ray_start.bat'):
            with subprocess.Popen(f'{base_dir}/scripts/ray_start.bat') as p:
                p.wait()
            initialize_ray(src_dir, ray_config_file_path, ray_config, success_callback)
            return syncer
        else:
            raise


def initialize_ray(src_dir, ray_config_file_path, ray_config=None, success_callback=None):
    if ray.is_initialized():
        return
    if ray_config is None:
        ray_config = load_ray_config(ray_config_file_path)
    # runtime_env is required for cloudpickle to be able to find modules
    runtime_env = {'working_dir': src_dir}
    ray_head_hostname_or_ip_address = ray_config['ray_head']['hostname_or_ip_address']
    ray_head_client_server_port = ray_config['ray_head']['client_server_port']
    ray.init(address=f'ray://{ray_head_hostname_or_ip_address}:{ray_head_client_server_port}', runtime_env=runtime_env)
    monkey_patch_trainable_util_to_fix_checkpoint_paths()
    if success_callback is not None:
        success_callback()


def load_ray_config(config_file_path):
    if not os.path.exists(config_file_path):
        raise FileNotFoundError(f'config file not found at {config_file_path}')
    with open(config_file_path, 'r') as f:
        ray_config = yaml.safe_load(f)
    return ray_config


def clean_trial_results_dir(syncer, trial_results_dir):
    logger.info('cleaning local trial results dir...')
    trial_results_dir = os.path.expanduser(trial_results_dir)
    if os.path.exists(trial_results_dir):
        delete_dir_contents(trial_results_dir)
    else:
        os.makedirs(trial_results_dir, exist_ok=True)
    if syncer is not None:
        logger.info('cleaning ray worker trial results dir...')
        syncer.sync_from_driver_to_ray_worker()


def configure_remote_ray_runtime_environment(base_dir,
                                             driver_private_key_file,
                                             worker_user,
                                             worker_hostname_or_ip_address,
                                             worker_ssh_port,
                                             worker_platform,
                                             worker_setup_commands):
    worker_base_dir = expand_user_home_path(base_dir, worker_user, worker_platform)
    sync_cmd = f'rsync -avz -e "ssh -i {driver_private_key_file} -o StrictHostKeyChecking=no -p {worker_ssh_port}" --include="Pipfile" --include="requirements.txt" --exclude="*" {base_dir}/ {worker_user}@{worker_hostname_or_ip_address}:{worker_base_dir}/'
    logger.info(f'copying runtime environment configuration files to remote Ray runtime with command "{sync_cmd}"')
    try:
        output = str(subprocess.check_output(sync_cmd, shell=True)).replace('\\n', '\n')
        logger.info(output)
    except subprocess.CalledProcessError as e:
        logger.error(f'error copying runtime environment configuration files to remote Ray runtime with command "{sync_cmd}": {e}')

    if worker_setup_commands is not None and len(worker_setup_commands) > 0:
        setup_commands = ' && '.join(worker_setup_commands)
        logger.info('configuring remote Ray runtime environment...')
        configure_cmd = f'ssh -i {driver_private_key_file} -o StrictHostKeyChecking=no -p {worker_ssh_port} {worker_user}@{worker_hostname_or_ip_address} "{setup_commands}"'
        try:
            logger.info(f'configuring remote Ray runtime environment with command "{setup_commands}"')
            output = str(subprocess.check_output(configure_cmd, shell=True)).replace('\\n', '\n')
            logger.info(output)
        except subprocess.CalledProcessError as e:
            logger.error(f'error configuring remote Ray runtime environment with command "{setup_commands}": {e}')


def delete_dir_contents(dir_path):
    for filename in os.listdir(dir_path):
        file_path = os.path.join(dir_path, filename)
        if os.path.isfile(file_path) or os.path.islink(file_path):
            os.unlink(file_path)
        elif os.path.isdir(file_path):
            shutil.rmtree(file_path, ignore_errors=True)



from abc import ABC
from ray.air import CheckpointConfig, RunConfig
from ray.train.huggingface import HuggingFaceTrainer
from ray.train.torch import TorchConfig
import torch
from transformers import Trainer, TrainingArguments, WEIGHTS_NAME

from config import RUNS_DIR, config
from log import LOGS_DIR, log
from training.trainer_initializer_base import TrainerInitializerBase
from ray_quickstart.util import platform


class HuggingFaceTrainerInitializerBase(TrainerInitializerBase, ABC):

    def trainer_args_init(self,
                          model,
                          tensorboard_logging_strategy='epoch', # 'no', 'epoch', or 'steps'
                          save_strategy='epoch', # 'no', 'epoch', or 'steps'
                          evaluation_strategy='epoch',
                          disable_tqdm=False): # 'no', 'epoch', or 'steps'
        if self.config.get_run_on_ray_cluster():
            load_best_model_at_end = False
        else:
            load_best_model_at_end=True
        args = TrainingArguments(
            seed=self.config.seed,
            data_seed=self.config.seed,
            remove_unused_columns=True,
            output_dir=f'{RUNS_DIR}/{model.model_name}',
            logging_dir=f'{LOGS_DIR}/tensorboard', # logging for TensorBoard
            logging_strategy=tensorboard_logging_strategy,
            evaluation_strategy=evaluation_strategy,
            save_strategy=save_strategy,
            load_best_model_at_end=load_best_model_at_end,
            optim='adamw_torch',
            use_mps_device=platform.is_mac() and not self.config.get_run_on_ray_cluster() and self.config.device_type == 'mps',
            disable_tqdm=disable_tqdm,
        )
        train_params = model.load_training_params()
        if train_params is not None:
            for key, value in train_params.items():
                setattr(args, key, value)
        return args

    def trainer_init(self, model, args, train_dataset, eval_dataset, scaling_config):
        data_collator = self.data_collator_init(model)
        trainer = HuggingFaceTrainer(
            trainer_init_per_worker=self.trainer_init_per_worker,
            scaling_config=scaling_config,
            datasets={'train': train_dataset, 'evaluation': eval_dataset},
            trainer_init_config={'model': model,
                                 'args': args,
                                 'data_collator': data_collator,
                                 'compute_metrics': self.compute_metrics_init()},
            torch_config=TorchConfig(backend='gloo'),
            run_config=RunConfig(name=model.model_name,
                                 local_dir=config.trial_results_dir,
                                 checkpoint_config=CheckpointConfig(num_to_keep=3),
                                 log_to_file=f'{model.model_name}.log')
        )
        return trainer

    def trainer_init_per_worker(self, train_dataset, eval_dataset, **trainer_init_config):
        model = 'model' in trainer_init_config and trainer_init_config['model'] or None
        model_init = 'model_init' in trainer_init_config and trainer_init_config['model_init'] or None
        args = trainer_init_config['args']
        # reinitialize the training arguments so that it gets initialized correctly on the worker
        args = TrainingArguments(
            seed=args.seed,
            data_seed=args.seed,
            remove_unused_columns=args.remove_unused_columns,
            output_dir=args.output_dir,
            overwrite_output_dir=args.overwrite_output_dir,
            learning_rate=args.learning_rate,
            per_device_train_batch_size=args.per_device_train_batch_size,
            per_device_eval_batch_size=args.per_device_eval_batch_size,
            num_train_epochs=args.num_train_epochs,
            weight_decay=args.weight_decay,
            logging_dir=args.logging_dir, # logging for TensorBoard
            logging_strategy=args.logging_strategy,
            evaluation_strategy=args.evaluation_strategy,
            save_strategy=args.save_strategy,
            load_best_model_at_end=args.load_best_model_at_end,
            optim=args.optim,
            #max_steps=5,
            #no_cuda=True,
            use_mps_device=args.use_mps_device,
            disable_tqdm=args.disable_tqdm,
        )
        data_collator = trainer_init_config['data_collator']
        compute_metrics = 'compute_metrics' in trainer_init_config and trainer_init_config['compute_metrics'] or None
        return Trainer(
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            model=model,
            model_init=model_init,
            args=args,
            data_collator=data_collator,
            compute_metrics=compute_metrics
        )

    def update_model_with_best_checkpoint(self, model, config, checkpoints, default_eval_metric):
        if checkpoints is None or len(checkpoints) == 0:
            return
        best_checkpoint_index = None
        best_checkpoint_value = None
        train_params = model.load_training_params()
        if train_params is not None and 'metric_for_best_model' in train_params:
            eval_metric = train_params['metric_for_best_model']
        else:
            eval_metric = default_eval_metric
        for index, checkpoint_info in enumerate(checkpoints):
            checkpoint_metrics = checkpoint_info[1]
            metric_value = checkpoint_metrics[eval_metric]
            if best_checkpoint_value is None or metric_value > best_checkpoint_value:
                best_checkpoint_value = metric_value
                best_checkpoint_index = index
        best_checkpoint = checkpoints[best_checkpoint_index][0]
        best_checkpoint_metrics = checkpoints[best_checkpoint_index][1]
        log.info(f'Best checkpoint metrics: {best_checkpoint_metrics}')
        log.info(f'Best checkpoint {eval_metric}: {best_checkpoint_value}')

        checkpoint_path = best_checkpoint.uri[7:]
        state_dict = torch.load(f'{checkpoint_path}/{WEIGHTS_NAME}', map_location=config.device_type)
        model.load_state_dict(state_dict)
        model.save_model()
